-------------------------------------------------------------------------------
-- Copyright (c) 2011 Sierra Wireless and others.
-- All rights reserved. This program and the accompanying materials
-- are made available under the terms of the Eclipse Public License v1.0
-- which accompanies this distribution, and is available at
-- http://www.eclipse.org/legal/epl-v10.html
--
-- Contributors:
--     Sierra Wireless - initial API and implementation
-------------------------------------------------------------------------------
-- Runtime assistance bootstrapping code
-------------------------------------------------------------------------------

module(..., package.seeall)

local RPC_PORT = os.getenv("LUA_INSPECTOR_PORT") or 12333

do
    -- table containing all data about loaded module, application itself 
    -- is considered as a module and is the entry point for building
    -- dependency tree
    required_modules = { ["<app>"] = { deps = {}, loaded_by = {} } }
    local raw_require = require
    local getinfo = debug.getinfo
    -- table used as stack to get module dependencies during loading
    local dependency_stack = { required_modules["<app>"].deps }
    
    function _G.require(modname, ...)
        -- gets or create deta table for current module
        local modtable = required_modules[modname] or { deps = {}, loaded_by = {} }
        required_modules[modname] = modtable
        
        local info = getinfo(2, "Sl")
        -- if called by copcall, getinfo cannot retrieve info from caller (another thread)
        local usable_info = info and { file = info["source"], line = info["currentline"] } 
                                          or { file = "<unknown>", line = -1 }
        table.insert(modtable.loaded_by, usable_info)
        -- add module as dependency of currenly loaded module (on the top of loading stack)
        table.insert(dependency_stack[#dependency_stack], modname)
        
        -- push the loaded module as current module in the stack and load it
        table.insert(dependency_stack, modtable.deps)
        local results = {raw_require(modname, ...)}
        table.remove(dependency_stack)
        
        return unpack(results)
    end
end

local utils = require"utils"
local rpc = require"rpc"
local sched = require"sched"
local log = require"log"
local dump_pool = require"debugintrospection"


do
    local threads = setmetatable({}, { __mode = "k" })
    local old_create = coroutine.create
    function coroutine.create(...)
        local thread = old_create(...)
        threads[thread] = true
        return thread
    end
end



------------------------------------------------------------------------
-- RPC Data dumping
------------------------------------------------------------------------

-- RPC connection handling
--------------------------

-- The called RPC have the following prototype
--   dump(lag, data)
-- where lag is the number of seconds between dump and actual send (if there is no connection when data is dumped) and data is the dumped data

-- Only one IDE can be connected at once, there is no particular configuration to do on Lua side, except setting RPC port on which
-- server will listen.
-- If dumps are done while there is not any connected IDE, they are stored and will be sent on the next time that an IDE is connected.
local ide_sock

-- Queue of all pending dumps that cannot be sent because of lack of IDE connection.
-- When a dump is stored in this table, it has the form { dump, timestamp } where timestamp is the result of os.time() when the dump is done. 
local pending_dumps = { max = 10 }

--- Sends a dump through RCP if possible. If there is no active connection, dump will be stored and sent when possible.
-- @param dump (table) dump to send: instance of debugintrospection (can be stored)
-- @param timestamp (number or nil) result of os.time() when the dump is done. Used internally, for external calls, just ignore this.
local function send_dump(dump, timestamp)
    -- try to send the dump
    if ide_sock then
        local _, err = ide_sock:call("dump", timestamp and os.difftime(os.time(), timestamp) or 0, dump.dump)
        if err then
            log("INSPECT", "ERROR", "Failed to send dump to ide: %s", err)
            ide_sock = nil
        end
    end
    
    -- if there is no connection, or if it has been closed, save the dump for further send
    if not ide_sock then
        while #pending_dumps >= pending_dumps.max do table.remove(pending_dumps, 1) end
        pending_dumps[#pending_dumps+1] = {dump, timestamp or os.time()}
    end
end


-- background task which constantly listen for a new IDE to connect with. If a new connection happend, the old one is replaced.
sched.run(function()
    local failures = 0
    while failures < 5 do -- failure control: after 5 successive failures, the service is stopped.
        local client, err = rpc.newserver("*", RPC_PORT)
        if err then
            log("INSPECT", "ERROR", "Failed to open server socket: %s", err)
            failures = failures + 1
            sched.wait(1)
        else
            failures = 0
            ide_sock = client
            -- send all pending dumps with the new connection, pending_dumps is immediately set to a new table to avoid
            -- weird problems if the new connection fails as soon as it's established
            local dumps = pending_dumps
            pending_dumps = { max = dumps.max }
            for _, dump in ipairs(dumps) do
                send_dump(unpack(dump))
            end
        end
    end
    log("INSPECT", "ERROR", "Too many server failures, stopping.")
end)


-- Public API
-------------

local dump_config = nil -- defaults to dump_table default

--- Dumps given values to dump registry. 
-- This is the rough equivalent of <code>dump_table({...}, what)</code> where <code>what</code> is the value given to set_config.
-- The only special case is when ... contains a single table: it is not wrapped in a new table.
-- @see set_config
-- @see dump_table
function dump(...)
    local value
    if select("#", ...) == 1 and type(...) == "table" then
        value = (...)
    else
        value = {...}
    end
    dump_table(value, dump_config)
end

--- Sets the switches of <code>dump</code> function. Default value is tha same as <code>dump_table</code> default.
-- @param what (string) Dump switches, see <code>dump_table</code> for description.
-- @see dump
-- @see dump_table
function set_config(what)
    dump_config = what
end

--- Dumps given table in dump registry.
-- The dump can be configured to avoid dumping useless values. This configuration is done like debug.getinfo with the following switches :
--   * "m" dumps metatables
--   * "u" dumps upvalues
--   * "l" dumps local values
--   * "e" dumps function environments
--   * "s" dumps thread stacks
-- @param t (table) table to dump
-- @param what (string) switches described above, defaults to "ul"
function dump_table(t, what)
    assert(type(t) == "table", "Value to dump must be a table")
    what = what or "ul"
    local context = dump_pool:new(
        what:find("l", 1, true),
        what:find("u", 1, true),
        what:find("m", 1, true),
        what:find("s", 1, true),
        what:find("e", 1, true))
    context.dump.root = context:table(t)
    -- The bare minimal information is dumped about caller function
    context.dump_locales, context.dump_upvalues, context.dump_fenv = false, false, false
    context.dump.caller = context["function"](context, 3)
    send_dump(context)
end

--- Entry point for remote RPC calls, expr is a string which, once evaluated,
-- must return a table. This table is then dumped and returned.
-- Errors are NOT catched, so this function propagates error if expr fails or 
-- does not result in a table.
-- If expr is nil, dumped table defaults to global table.
function dump_expr(expr)
    -- debug hooks are disabled during the dump because performance loss is too high
    local hook, mask, count = debug.gethook()
    debug.sethook()
    
    local success, result = pcall(function()
        local t
        if expr then
            local res = {assert(loadstring("return "..expr))()}
            -- if function has returned a single table, drop wrapping table
            t = (#res == 1 and type(res[1]) == "table") and res[1] or res
        else
            t = getfenv(0) -- defaults to global env
        end
        
        local context = dump_pool:new(true, true, true, true, true)
        context.dump.root = context:table(t)
        return context.dump
    end)
    
    debug.sethook(hook, mask, count)
    if success then return result else error(result) end
end
